#!/usr/bin/env python
#
# GABI - Genome Assembly by Bayesian Inference
#
# Copyright 2013-2014, Brown University, Providence, RI. All Rights Reserved.
#
# This file is part of GABI.
#
# GABI is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GABI is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GABI.  If not, see <http://www.gnu.org/licenses/>.

"""
Prototype of a Markov Chain Monte Carlo de novo sequence assembler.
"""

import os
import random

from time import time

import trace
from sampler import Sampler
from graph import Graph

from biolite import diagnostics
from biolite.pipeline import IlluminaPipeline

pipe = IlluminaPipeline('sample', __doc__)

# Command-line arguments.

pipe.add_path_arg('--graphml', '-g', help="""
    graphml file generated by the GABI debruijn script.""")

pipe.add_arg('--samples', '-n', type=int, metavar='N', default=10000, help="""
    Collect N samples.""")

pipe.add_arg('--perturbations', '-P', type=int, metavar='N', default=1, help="""
    Perform N perturbations when proposing a new assembly in each sample.""")

pipe.add_arg('--contigs', '-c', type=int, metavar='N', default=1, help="""
    Estimated number of contigs (e.g. choromosomes, independent genomes, etc.),
    used to calculate the prior probability of an assembly.""")

pipe.add_arg('--scale', '-k', type=int, metavar='K', default=10, help="""
    Scale parameter for the gamma distributions used for the priors.""")

pipe.add_arg('--nodata', action='store_true', default=False, help="""
    Turn off likelihood calculations and only use priors in MCMC.""")

pipe.add_arg('--flatprior', action='store_true', default=False, help="""
    Use a flat prior distribution.""")

pipe.add_arg('--seed', type=long, default=-1, help="""
    Deterministic (instead of random) sampling, starting from a seed value.""")

# Required arguments.

pipe.add_arg('--genome_size', '-s', type=int, metavar='SIZE', help="""
    Estimated size of the genome, used to calculate the prior probability of
    an assembly.""")

pipe.add_arg('--insert_mean', type=float, help="""
    Mean insert size of the paired read data.""")

pipe.add_arg('--insert_stdev', type=float, help="""
    St. dev. of the insert size of the paired read data.""")


# Pipeline stages.

@pipe.stage
def sample(
    data, graphml, genome_size, contigs, scale, nodata, flatprior, seed,
    insert_mean, insert_stdev, samples, perturbations, outdir):
    """
    Perform MCMC sampling: iteratively perturb the graph, re-calculate
    likelihood with LAP, and accept new states that pass the odds ratio.
    """

    if seed >= 0:
        random.seed(seed)

    graph = Graph(graphml)

    # Create a sampler that calculates likelihood and prior probabilities.
    sampler = Sampler(
                data[-1], genome_size, contigs, scale,
                insert_mean, insert_stdev, nodata, flatprior)

    # Store all visited states, to avoid duplicate calculations.
    state_cache = {}

    # Open the trace database and check for existing runs.
    trace_path = os.path.join(outdir, 'mcmc.trace.sqlite')
    last = trace.connect(trace_path)
    diagnostics.log_path(trace_path, 'trace')

    if last is None:
        # Random initial state.
        graph.randomize_state()
        # Start with an initial assembly from the unperturbed graph.
        current = sampler.sample(graph.contigs(), 'initial-contigs.fa')
        save_state = graph.get_state()
        trace.add(0, current, 1, 0, *save_state)
        start = 0
    else:
        # Start form an existing chain.
        i = trace.last_accepted()
        current = trace.get_sample(i)
        save_state = trace.get_state(i)
        graph.set_state(*save_state)
        start = last

    timer = time()
    # Perform MCMC.
    for i in xrange(start, start+samples):

        # Perturb graph.
        for _ in xrange(perturbations):
            graph.perturb_path()

        # Run assembler on perturbed graph and calculate new log-likelihood
        # and log-prior.
        state = graph.get_state()
        if state[0] in state_cache:
            candidate = state_cache[state[0]]
        else:
            candidate = sampler.sample(graph.contigs(), 'candidate-contigs.fa')
            state_cache[state[0]] = candidate

        # MCMC: calculate ratio of new and old posterior probabilities
        # and use Hastings ratio to decide whether to accept new assembly.
        accept = candidate.odds_ratio(current)

        # Restart timer.
        runtime = time() - timer
        timer = time()

        # Log the sample (before modifying accepting).
        trace.add(i+1, candidate, accept, runtime, *state)

        # Accept or reject the sample.
        if accept > 0:
            current = candidate
            save_state = state
        else:
            graph.set_state(*save_state)


if __name__ == '__main__':
    pipe.run()
    diagnostics.merge()

